package main

import (
	"net/url"
	"sync/atomic"
	"testing"
	"time"

	model "github.com/DataDog/agent-payload/process"
	"github.com/DataDog/datadog-agent/pkg/process/config"
	"github.com/DataDog/datadog-agent/pkg/process/util/api"
	"github.com/stretchr/testify/assert"
	"github.com/stretchr/testify/require"
)

func TestUpdateRTStatus(t *testing.T) {
	assert := assert.New(t)
	cfg := config.NewDefaultAgentConfig(false)
	c, err := NewCollector(cfg)
	assert.NoError(err)
	// XXX: Give the collector a big channel so it never blocks.
	c.rtIntervalCh = make(chan time.Duration, 1000)

	// Validate that we switch to real-time if only one response says so.
	statuses := []*model.CollectorStatus{
		{ActiveClients: 0, Interval: 2},
		{ActiveClients: 3, Interval: 2},
		{ActiveClients: 0, Interval: 2},
	}
	c.updateStatus(statuses)
	assert.Equal(int32(1), atomic.LoadInt32(&c.realTimeEnabled))

	// Validate that we stay that way
	statuses = []*model.CollectorStatus{
		{ActiveClients: 0, Interval: 2},
		{ActiveClients: 3, Interval: 2},
		{ActiveClients: 0, Interval: 2},
	}
	c.updateStatus(statuses)
	assert.Equal(int32(1), atomic.LoadInt32(&c.realTimeEnabled))

	// And that it can turn back off
	statuses = []*model.CollectorStatus{
		{ActiveClients: 0, Interval: 2},
		{ActiveClients: 0, Interval: 2},
		{ActiveClients: 0, Interval: 2},
	}
	c.updateStatus(statuses)
	assert.Equal(int32(0), atomic.LoadInt32(&c.realTimeEnabled))
}

func TestUpdateRTInterval(t *testing.T) {
	assert := assert.New(t)
	cfg := config.NewDefaultAgentConfig(false)
	c, err := NewCollector(cfg)
	assert.NoError(err)
	// XXX: Give the collector a big channel so it never blocks.
	c.rtIntervalCh = make(chan time.Duration, 1000)

	// Validate that we pick the largest interval.
	statuses := []*model.CollectorStatus{
		{ActiveClients: 0, Interval: 3},
		{ActiveClients: 3, Interval: 2},
		{ActiveClients: 0, Interval: 10},
	}
	c.updateStatus(statuses)
	assert.Equal(int32(1), atomic.LoadInt32(&c.realTimeEnabled))
	assert.Equal(10*time.Second, c.realTimeInterval)
}

func TestHasContainers(t *testing.T) {
	assert := assert.New(t)

	collectorProc := model.CollectorProc{}
	collectorContainer := model.CollectorContainer{}
	collectorRealTime := model.CollectorRealTime{}
	collectorContainerRealTime := model.CollectorContainerRealTime{}
	collectorConnections := model.CollectorConnections{}

	assert.Equal(0, getContainerCount(&collectorProc))
	assert.Equal(0, getContainerCount(&collectorContainer))
	assert.Equal(0, getContainerCount(&collectorRealTime))
	assert.Equal(0, getContainerCount(&collectorContainerRealTime))
	assert.Equal(0, getContainerCount(&collectorConnections))

	c := &model.Container{Type: "Docker"}
	cs, cs2 := &model.ContainerStat{Id: "1234"}, &model.ContainerStat{Id: "5678"}

	collectorProc.Containers = append(collectorProc.Containers, c)
	collectorContainer.Containers = append(collectorContainer.Containers, c)
	collectorRealTime.ContainerStats = append(collectorRealTime.ContainerStats, cs, cs2)
	collectorContainerRealTime.Stats = append(collectorContainerRealTime.Stats, cs)

	assert.Equal(1, getContainerCount(&collectorProc))
	assert.Equal(1, getContainerCount(&collectorContainer))
	assert.Equal(2, getContainerCount(&collectorRealTime))
	assert.Equal(1, getContainerCount(&collectorContainerRealTime))
}

func TestRemovePathIfPresent(t *testing.T) {
	for _, tt := range []struct {
		input    string
		expected string
	}{
		{input: "http://foo.com", expected: "http://foo.com"},
		{input: "http://foo.com/", expected: "http://foo.com"},
		{input: "http://foo.com/api/v1", expected: "http://foo.com"},
		{input: "http://foo.com?foo", expected: "http://foo.com"},
		{input: "http://foo.com/api/v1/?foo", expected: "http://foo.com"},
		{input: "http://foo.com/api/v1?foo", expected: "http://foo.com"},
		{input: "http://foo.com:8080", expected: "http://foo.com:8080"},
		{input: "http://foo.com:8080/", expected: "http://foo.com:8080"},
		{input: "http://foo.com:8080/api/v1", expected: "http://foo.com:8080"},
	} {
		u, err := url.Parse(tt.input)
		require.NoError(t, err)

		assert.Equal(t, tt.expected, removePathIfPresent(u))
	}
}

func TestKeysPerDomain(t *testing.T) {
	for _, tt := range []struct {
		input    []api.Endpoint
		expected map[string][]string
	}{
		{
			input: []api.Endpoint{
				{APIKey: "key1", Endpoint: getEndpoint(t, "http://foo.com")},
			},
			expected: map[string][]string{
				"http://foo.com": {"key1"},
			},
		},
		{
			input: []api.Endpoint{
				{APIKey: "key1", Endpoint: getEndpoint(t, "http://foo.com")},
				{APIKey: "key2", Endpoint: getEndpoint(t, "http://foo.com")},
			},
			expected: map[string][]string{
				"http://foo.com": {"key1", "key2"},
			},
		},
		{
			input: []api.Endpoint{
				{APIKey: "key1", Endpoint: getEndpoint(t, "http://foo.com")},
				{APIKey: "key2", Endpoint: getEndpoint(t, "http://bar.com")},
			},
			expected: map[string][]string{
				"http://foo.com": {"key1"},
				"http://bar.com": {"key2"},
			},
		},
		{
			input: []api.Endpoint{
				{APIKey: "key1", Endpoint: getEndpoint(t, "http://foo.com")},
				{APIKey: "key2", Endpoint: getEndpoint(t, "http://bar.com")},
				{APIKey: "key3", Endpoint: getEndpoint(t, "http://foo.com")},
			},
			expected: map[string][]string{
				"http://foo.com": {"key1", "key3"},
				"http://bar.com": {"key2"},
			},
		},
	} {
		assert.Equal(t, tt.expected, keysPerDomains(tt.input))
	}
}

func getEndpoint(t *testing.T, u string) *url.URL {
	e, err := url.Parse(u)
	assert.NoError(t, err)
	return e
}
